/*
   cfg.c -- PIKT config file routines

   $Header: /home/cvs/ucgsb/pikt/src/piktc/cfg.c,v 1.32 2000/05/14 23:56:08 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2000 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
*/

#include "../lib/common.h"
#include "piktc.h"

int firstalarm;

/* register a macro/alarm def component
 */
void
regcomponent (ident *idp, char *component)
{
	def *ndfp;

	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "regcomping COMP @%s@, ident is %s", component, idp->name);
	ndfp = (def *) mymalloc(sizeof(def), "regcomponent");
	ndfp->next = (def *) NULL;
	ndfp->component = component;
	if (! idp->definition)
		idp->definition = ndfp;
	else
		idp->defcurr->next = ndfp;
	idp->defcurr = ndfp;
	return;
}

/* begin writing the pikt .alt file
 */
void
beginalarms (char *alertname)
{
	static char _beginalarms_[] = "beginalarms";
	static char *outfile;
	nam *naltp;
	char confline[MAXLINE+1];

	outfile = (char *) mymalloc((strlen(stgdir)+strlen(alertname)+6) * sizeof(char), _beginalarms_);
	sprintf(outfile, "%s/%s.alt", stgdir, alertname);
	if ((cfgout = fopen(outfile, "w")) == NULL) {
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in beginalarms(), fopen() error for %s", outfile);
	}
	free(outfile);
	naltp = (nam *) mymalloc(sizeof(nam), _beginalarms_);
	naltp->next = (nam *) NULL;
	naltp->name = mystrdup(alertname, _beginalarms_);
	naltp->state = ADD;
	if (! alerts)
		alerts = naltp;
	else {
		naltp->next = alerts;
		alerts = naltp;
	}
	if (nicecmd)
		sprintf(confline, "%s %s %s %s %s %s %s %s/pikt", mins, hrs, doms, mons, dows, drift, nicecmd, bindir);
	else
		sprintf(confline, "%s %s %s %s %s %s %s/pikt", mins, hrs, doms, mons, dows, drift, bindir);
	if (mailcmd) {
		strcat(confline, " +M ");
		strcat(confline, "\"");
		strcat(confline, mailcmd);
		strcat(confline, "\"");
	}
	if (lpcmd) {
		strcat(confline, " +L ");
		strcat(confline, "\"");
		strcat(confline, lpcmd);
		strcat(confline, "\"");
	}
	strcat(confline, " +A ");
	strcat(confline, alertname);
	naltp->aux = mystrdup(confline, _beginalarms_);
	naltp->dir = (char *) NULL;
	naltp->file = (char *) NULL;
	naltp->mode = (char *) NULL;
	naltp->uid = (char *) NULL;
	naltp->gid = (char *) NULL;
	naltp->state = NONE;
	firstalarm = TRUE;
}

/* write an alarm definition to the pikt .alt file
 */
void
writealarm (char *alarmname)
{
	ident *idp;
	def *dfp;
	const char s1[] = " init begin rule end ";
	const char s2[] = " status level task input filter seps separators keys dat data delete do if elif elsif elseif else endif fi while endwhile repeat until for foreach endfor endforeach switch endswitch endsw breakswitch breaksw case default break cont continue again redo next skip last pause quit die output set unset exec ";
	static char *comp;

	idp = lookupident(alarmname, ALMCFG);
	if (idp) {
		if (firstalarm)
			firstalarm = FALSE;
		else
			fprintf(cfgout,"\n\n");
		fprintf(cfgout,"%s",alarmname);
		foreach (dfp, idp->definition) {
			comp = (char *) mymalloc(strlen(dfp->component)+3, "writealarm");
			sprintf(comp, " %s ", dfp->component);
			if (strstr(s1, comp))
				fprintf(cfgout, "\n\t%s ", dfp->component);
			else if (strstr(s2, comp) || ! strcmp(" for ( ",comp) || ! strcmp(" switch ( ",comp))
				fprintf(cfgout, "\n\t\t%s ", dfp->component);
			else if (! strcmp("[",dfp->component) || (dfp->next && (! strcmp("[",dfp->next->component) || ! strcmp("]",dfp->next->component))))
				fprintf(cfgout, "%s", dfp->component);
			else
				fprintf(cfgout, "%s ", dfp->component);
			free(comp);
		}
	} else
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in writealarm(), %s, %s alarm not registered", cfgfilecurr, alarmname);
	return; 
}

/* close the pikt .alt file
 */
void
endalarms (char *alert)
{

	fprintf(cfgout,"\n");
	if (fclose(cfgout))
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in endalarms(), fclose() error for %s/%s.alt", altdir, alert);
	return;
}

/* begin writing the other/program file
 */
void
beginfile (char *name, char *filepath, char *mode, char *uid, char *gid, nam **list)
{
	static char _beginfile_[] = "beginfile";
	char *cp;
	char *dir, *file;
	static char *outfile;
	static char *diffile;
	nam *nfilp;

	if ((cp = strrchr(filepath, '/'))) {
		*cp = '\0';
		dir = mystrdup(filepath, _beginfile_);
		file = mystrdup(cp+1, _beginfile_);
		outfile = (char *) mymalloc((strlen(stgdir)+strlen(dir)+strlen(file)+2) * sizeof(char), _beginfile_);
		sprintf(outfile, "%s%s/%s", stgdir, dir, file);
		/* make the staging subdirs for this file */
		mkdirtree(outfile);
		if (isdiff) {
			diffile = (char *) mymalloc((strlen(difdir)+strlen(dir)+strlen(file)+2) * sizeof(char), _beginfile_);
			sprintf(diffile, "%s%s/%s", difdir, dir, file);
			/* make the diffing subdirs for this file */
			mkdirtree(diffile);
			free(diffile);
		}
	} else {
		dir = mystrdup("", _beginfile_);
		file = mystrdup(filepath, _beginfile_);
		outfile = (char *) mymalloc((strlen(stgdir)+strlen(file)+2) * sizeof(char), _beginfile_);
		sprintf(outfile, "%s/%s", stgdir, file);
	}
	if ((cfgout = fopen(outfile, "w")) == NULL)
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in beginfile(), fopen() error for %s", outfile);
	free(outfile);
	nfilp = (nam *) mymalloc(sizeof(nam), _beginfile_);
	nfilp->next = (nam *) NULL;
	if (*name == '/') {
		*cp = '/';	/* restore the '/' we blotted out above */
		if (strcmp(name, filepath))
			logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in beginalarm(), ident %s doesn't match filepath %s", name, filepath);
		else
			nfilp->name = mystrdup(file, _beginfile_);
	} else {
		nfilp->name = mystrdup(name, _beginfile_);
	}
	nfilp->aux = (char *) NULL;
	nfilp->dir = dir;
	nfilp->file = file;
	nfilp->mode = mode;
	nfilp->uid = uid;
	nfilp->gid = gid;
	nfilp->state = ADD;
	if (*list)
		nfilp->next = *list;
	*list = nfilp;
	return;
}

/* write a single line to the other/program file
 */
void
writefile (char *line)
{

	fprintf(cfgout, "%s\n", line);
	return;
}

/* close the other/program file
 */
void
endfile (char *file)
{
	char *cp;

	if ((cp = strrchr(file, '/')))
		cp = cp+1;
	else
		cp = file;
	if (fclose(cfgout))
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in endfile(), fclose() error for %s/%s", stgdir, cp);
	return;
}

/* begin writing the objectfile
 */
void
beginobjects (char *objectname)
{
	static char _beginobjects_[] = "beginobjects";
	static char *outfile;
	nam *nobjp;

	outfile = (char *) mymalloc((strlen(stgdir)+strlen(objectname)+6) * sizeof(char), _beginobjects_);
	sprintf(outfile, "%s/%s.obj", stgdir, objectname);
	if ((cfgout = fopen(outfile, "w")) == NULL)
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in beginobjects, fopen() error for %s", outfile);
	free(outfile);
	nobjp = (nam *) mymalloc(sizeof(nam), _beginobjects_);
	nobjp->next = (nam *) NULL;
	nobjp->name = mystrdup(objectname, _beginobjects_);
	nobjp->aux = (char *) NULL;
	nobjp->dir = (char *) NULL;
	nobjp->file = (char *) NULL;
	nobjp->mode = (char *) NULL;
	nobjp->uid = (char *) NULL;
	nobjp->gid = (char *) NULL;
	nobjp->state = ADD;
	if (! objects)
		objects = nobjp;
	else {
		nobjp->next = objects;
		objects = nobjp;
	}

}

/* write a single object to the objectfile
 */
void
writeobject (char *object)
{

	fprintf(cfgout,"%s\n",object);
}

/* close the objectfile
 */
void
endobjects (char *objs)
{

	if (fclose(cfgout))
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in endobjects(), fclose error for %s/%s.obj", objdir, objs);
	return;
}

/* return the length of the piktd.conf line for the given alert
 */
int
conflinelen (char *alertname)
{
	int len;

	len = strlen(pikt)+1;			/* e.g., '/pikt/bin/pikt ' */
	len += strlen(mins) + 1 +		/* 'x ' */
	       strlen(hrs) + 1 +		/* 'x ' */
	       strlen(doms) + 1 +		/* 'x ' */
	       strlen(mons) + 1 +		/* 'x ' */
	       strlen(dows) + 1 ;		/* 'x ' */
	if (nicecmd)
		len += 1 + strlen(nicecmd);
	if (mailcmd)
		len += 6 + strlen(mailcmd);	/* '+M "..." ' */
	if (lpcmd)
		len += 6 + strlen(lpcmd);	/* '+P "..." ' */
	len += 4 + strlen(alertname);		/* '+A ...\n' */
	return len;
}

/* given a pathname, make (as necessary) all directories in that path
 */
void
mkdirtree (char *path)
{
	register int i;
	register char c;

	i = 0;
	if (*(path+i++) == '/') {
		while ((c = *(path+i))) {
			if (c == '/') {
				*(path+i) = '\0';
				if (! isfile(path, FALSE)) {
					if (mkdir(path, (mode_t)(S_IRWXU|S_IRGRP|S_IWGRP)) == -1)	/* 750 */
						logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in mkdirtree(), mkdir() failure for %s", path);
				}
				*(path+i) = '/';
			}
			i++;
		}
	} else {
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in mkdirtree(), path %s is not an absolute path", path);
	}
	return;
}
