/*
   ident.c -- PIKT pikct ident(ifier) routines

   $Header: /home/cvs/ucgsb/pikt/src/piktc/ident.c,v 1.29 2000/03/04 16:29:55 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2000 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
 */

#include "../lib/common.h"
#include "piktc.h"

ident *idents[256+OBJCFG];
const char *identypes[] = { "operating system",
                            "host",
                            "hostgroup",
                            "hostalias",
                            "define",
                            "macro",
                            "alarm",
                            "alert",
                            "program",
                            "file",
                            "object"
                          };

unsigned char hash(unsigned char *, int);

/* do a lookup of an identifier in the ident list
 */
ident *
lookupident (char *identname, int cfgtype)
{
	ident *idp;

	foreach (idp, idents[hash(identname, cfgtype)]) {
		if (isdebug)
			logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "in lookupident(), strcmping %s with %s", identname, idp->name);
		if (! strcmp(identname,idp->name))
			return idp;
	}
	return (ident *) NULL;
}

/* register an identifier in the ident list
 */
ident *
regident (char *identname, int identype, int setdefbit)
{
	static char _regident_[] = "regident";
	ident *idp, *nidp, *aidp;
	int cfgtype;
	static unsigned long *tbs;

	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "regidenting IDENT @%s@", identname);
	/* map the identype to its corresponding cfgtype before doing the
	   lookup:
		OPS	->	SYSCFG
		HST	->	SYSCFG
		GRP	->	SYSCFG
		ALS	->	SYSCFG
		DEF	->	DEFCFG
		MCR	->	MCRCFG
		ALM	->	ALMCFG
		ALT	->	ALTCFG
		PRG	->	PRGCFG
		FIL	->	FILCFG
		OBJ	->	OBJCFG
	 */
	cfgtype = (identype <= ALS ? SYSCFG : identype - 3);
	if (lookupident(identname, cfgtype))
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regident(), %s already registered as %s", identname, identypes[identype]);
	nidp = (ident *) mymalloc(sizeof(ident), _regident_);
	nidp->next = (ident *) NULL;
	nidp->name = identname;
	nidp->type = identype;
	nidp->alias = (char *) NULL;
	nidp->definition = (def *) NULL;
	nidp->defcurr = (def *) NULL;
	nidp->prmnum = 0;
	nidp->called = FALSE;
	switch (identype) {
		case HST:	/* host */
			regname(&cfghosts, identname, FALSE, FALSE);	/* for command-line */
			*(hosts+hostnum++) = identname;	/* remember hostname */
			nidp->bs = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), _regident_);
			*(nidp->bs+((hostnum-1)/(8*sizeof(unsigned long)))) = (1 << ((hostnum-1) % (8*sizeof(unsigned long))));
			tbs = hbsfull;
			hbsfull = bsunion(hbsfull, nidp->bs, hbslongs);
			free(tbs);
			break;
		case ALS:	/* host alias */
			if ((aidp = lookupident(hostcurr, SYSCFG))) {
				if (aidp->type != HST) {
					fprintf(stderr, "alias %s is not a host\n", aidp->name);
					fflush(stderr);
					bailout = TRUE;
					return nidp;
				} else {
					nidp->alias = mystrdup(aidp->name, _regident_);
					nidp->bs = bsdup(aidp->bs, hbslongs);
				}
			} else {
				fprintf(stderr, "unknown host %s\n", hostcurr);
				fflush(stderr);
				bailout = TRUE;
				return nidp;;
			}
			break;
		case OPS:	/* operating system */
			*(opsys+opsysnum++) = identname;	/* remember os name */
			nidp->bs = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), _regident_);
			ghbs = nidp->bs;
			break;
		case GRP:	/* hostgroup */
			nidp->bs = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), _regident_);
			ghbs = nidp->bs;
			break;
		case DEF:	/* define */
			defnum++;
			nidp->bs = (unsigned long *) mycalloc(dbslongs, sizeof(unsigned long), _regident_);
			*(nidp->bs+((defnum-1)/(8*sizeof(unsigned long)))) = (1 << ((defnum-1) % (8*sizeof(unsigned long))));
			if (setdefbit) {	/* set the def bit */
				tbs = gdbs;
				gdbs = bsunion(gdbs, nidp->bs, dbslongs);
				free(tbs);
			} else {		/* set the undef bit */
				tbs = gubs;
				gubs = bsunion(gubs, nidp->bs, dbslongs);
				free(tbs);
			}
			/* dbsfull currently serves no purpose
			tbs = dbsfull;
			dbsfull = bsunion(dbsfull, nidp->bs, dbslongs);
			free(tbs);
			 */
			break;
		case MCR:	/* macro */
		case ALM:	/* alarm */
			/* nidp->definition = (def *) NULL; */
			break;
		default:	/* ALT/PRG/FIL/OBJ--alert/program/file/object */
			break;
	}
	for (idp=idents[hash(identname, cfgtype)]; idp && idp->next; idp=idp->next) ;
	if (! idp)
		idents[hash(identname, cfgtype)] = nidp;
	else
		idp->next = nidp;
	return nidp;
}

/* garbage collect all processed idents DEF and higher (i.e., everything
   except what's in systems.cfg)
 */
void
cleanup (void)
{
	register int i;
	static ident *idp, *tidp;
	static def *dfp, *tdfp;

	for (i=0; i<=255+OBJCFG; i++) {
		if (! idents[i])
			continue;
		idp = tidp = idents[i];
		while (idp && (idp->type < DEF)) {
			tidp = idp;
			idp = idp->next;
		}
		if (tidp != idp)
			tidp->next = (ident *) NULL;
		if (idp == idents[i])
			idents[i] = (ident *) NULL;
		while (idp) {
			tidp = idp;
			idp = idp->next;
			if (tidp->name) {
				if (isdebug)
					logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "freeing IDENT %s", tidp->name);
				free(tidp->name);
			}
			if (tidp->alias) {
				if (isdebug)
					logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "freeing IDENT alias %s", tidp->alias);
				free(tidp->alias);
			}
			dfp = tidp->definition;
			while (dfp) {
				tdfp = dfp;
				dfp = dfp->next;
				if (tdfp->component) {
					if (isdebug)
						logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "freeing COMP %s", tdfp->component);
					free(tdfp->component);
				}
				free(tdfp);
			}
			free(tidp);
		}
	}
	return;
}

/* sum the ident's first and last chars and add the cfgtype (to allow for
   multiple idents instances in the hash table, but just one per config file
   type)
 */
unsigned char
hash (unsigned char *idnt, int cfgtype)
{
	unsigned char c1, c2, c;
	register int i;

	i = 0;
	c1 = *idnt;
	while (*(idnt+i++)) ;
	c2 = *(idnt+i-2);
	c = c1 + c2 + cfgtype;
	if (c > 255+OBJCFG) {
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in hash(), for ident %s, invalid hash val %d", idnt, c);
	}
	return c;
}

/* make the built-in, predefined pikt macros
 */
void
makepiktmacros (void)
{
	static char _makepiktmacros_[] = "makepiktmacros";
	ident *idp;
	register int i;

	idp = regident(mystrdup("pikthostname", _makepiktmacros_), MCR, FALSE);
	regcomponent(idp, mystrdup(hostcurr, _makepiktmacros_));

	/* we have to do these again for each hostcurr,
	   since macros are eliminated in each cleanup()
	 */

	idp = regident(mystrdup("piktdir", _makepiktmacros_), MCR, FALSE);
	regcomponent(idp, mystrdup(homdir, _makepiktmacros_));

	idp = regident(mystrdup("piktmaster", _makepiktmacros_), MCR, FALSE);
	regcomponent(idp, mystrdup(master, _makepiktmacros_));

	idp = regident(mystrdup("pikthosts", _makepiktmacros_), MCR, FALSE);
	for (i=0; i<hostnum; i++)
		regcomponent(idp, mystrdup(*(hosts+i), _makepiktmacros_));

	idp = regident(mystrdup("piktopsystems", _makepiktmacros_), MCR, FALSE);
	for (i=0; i<opsysnum; i++)
		regcomponent(idp, mystrdup(*(opsys+i), _makepiktmacros_));

	idp = regident(mystrdup("piktnever", _makepiktmacros_), MCR, FALSE);
	regcomponent(idp, mystrdup(PIKTNEVER, _makepiktmacros_));

	idp = regident(mystrdup("piktnullchar", _makepiktmacros_), MCR, FALSE);
	regcomponent(idp, mystrdup(piktnullchar, _makepiktmacros_));

	idp = regident(mystrdup("_", _makepiktmacros_), MCR, FALSE);
	regcomponent(idp, mystrdup(piktnullchar, _makepiktmacros_));

	idp = regident(mystrdup("piktspacechar", _makepiktmacros_), MCR, FALSE);
	regcomponent(idp, mystrdup(piktspacechar, _makepiktmacros_));

	idp = regident(mystrdup("pikttabchar", _makepiktmacros_), MCR, FALSE);
	regcomponent(idp, mystrdup(pikttabchar, _makepiktmacros_));

	idp = regident(mystrdup("piktnewlinechar", _makepiktmacros_), MCR, FALSE);
	regcomponent(idp, mystrdup(piktnewlinechar, _makepiktmacros_));

	return;
}

/* set path to the default "/bin/nice" (which is accurate for most
   systems); if nice has been defined as a macro, if that macrodef is
   a filepath, and if that file exists, then set path to that filepath
   instead; combine with prio to make nicecmd
 */
char *
getnicecmd (char *prio)
{
	ident *idp;
	struct stat fbuf;
	const char *path;
	char cmd[MAXCMD+1];
	char *r;

	path = "/bin/nice";	/* default path */
	if ((idp = lookupident("nice", MCRCFG))) {
		if (lstat(idp->definition->component, &fbuf) == 0)
			path = idp->definition->component;
	}
	sprintf(cmd, "%s -%s", path, prio);
	r = mystrdup(cmd, "getnicecmd");
	return r;
}
