%{

/*
   alt.l -- PIKT alarm script lexer

   $Header: /home/cvs/ucgsb/pikt/src/piktd/alt.l,v 1.35 2000/06/08 19:11:29 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2000 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA

   ----------------------------------------------------------------------------

   this is the sausage factory; not for the squeamish; enter at your own risk!

   on our test irix system, flex failed to match regexps of the form
   '{D}{x,y}', which led to parser failures; hence substitutions of the
   form '{D}+' below

 */

#include "../lib/common.h"
#include "piktd.h"
#include "alt.tab.h"

char _altlex_[] = "altlex";

#define RETURN(token) altlval.sval = mystrdup(yytext, _altlex_); return token

/* internally, every Pikt value is considered an array element; in the case
   of scalars, we pretend as if $foo were really written as $foo[0]
 */
#define FAKEARRAY unput(']'); unput('0'); unput('['); unput('['); unput('[')

int ii, jj, kk;
int d;
char cc;
char *s1, *s2;
char *yyt;
int yyl;

int context;	/* STRING/NUMBER */

char outstr[MAXBUF+1];

char *nam;

int fn = 0;

int isarray;
int isfakearray;

void alterror(const char *);

%}

W		[ \t]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]
DLU		[0-9A-Za-z_]

%x CS DS SS FS

%option noyywrap

%%

<*>^#.+\n			{ alterror("syntax error");
				}

<*>^{L}[-_\.0-9A-Za-z]*		{ context = NIL;
				  BEGIN(INITIAL);
				  RETURN(_STANZA);
				}

<INITIAL,DS,SS>status		{ return _STATUS; }
<INITIAL,DS,SS>active		{ return _ACTIVE; }
<INITIAL,DS,SS>inactive		{ return _INACTIVE; }

<INITIAL,DS,SS>level		{ return _LEVEL; }
<INITIAL,DS,SS>emerg|emergency	{ return _EMERG; }
<INITIAL,DS,SS>alert|urgent	{ return _ALERT; }
<INITIAL,DS,SS>crit|critical	{ return _CRIT; }
<INITIAL,DS,SS>err|error	{ return _ERR; }
<INITIAL,DS,SS>warn|warning	{ return _WARNING; }
<INITIAL,DS,SS>notice		{ return _NOTICE; }
<INITIAL,DS,SS>info		{ return _INFO; }
<INITIAL,DS,SS>debug		{ return _DEBUG; }

<INITIAL,DS,SS>task		{ return _TASK; }

<INITIAL,DS,SS>input		{ return _INPUT; }
<INITIAL,DS,SS>proc|process	{ BEGIN(CS);
				  return _PROCESS;
				}
<INITIAL,DS,SS>file		{ return _REGFILE; }
<INITIAL,DS,SS>logfile		{ return _LOGFILE; }

<INITIAL,DS,SS>filter		{ BEGIN(CS);
				  return _FILTER;
				}

<INITIAL,DS,SS>seps|separators	{ return _SEPARATORS; }
<INITIAL,DS,SS>dat|data		{ BEGIN(DS);
				  return _DAT;
				}
<INITIAL,DS,SS>keys		{ BEGIN(DS);
				  return _KEYS;
				}

<INITIAL,DS,SS>init		{ return _INIT; }
<INITIAL,DS,SS,CS>begin		{ BEGIN(INITIAL);
				  return _BEGIN;
				}
<INITIAL,DS,SS,CS>end		{ BEGIN(INITIAL);
				  return _END;
				}
<INITIAL,DS,SS,CS>rule		{ BEGIN(INITIAL);
				  return _RULE;
				}

<INITIAL,DS,SS>do		{ context = NIL;
				  return _DO;
				}
<INITIAL,DS,SS>if		{ context = NUMBER;
				  return _IF;
				}
<INITIAL,DS,SS>if{W}*"("	{ context = NUMBER;
				  unput('[');
				  return _IF;
				}
<INITIAL,DS,SS>(elif|elsif|elseif) {
				  context = NUMBER;
				  return _ELIF;
				}
<INITIAL,DS,SS>(elif|elsif|elseif){W}*"(" {
				  context = NUMBER;
				  unput('[');
				  return _ELIF;
				}
<INITIAL,DS,SS>else		{ context = NIL;
				  return _ELSE;
				}
<INITIAL,DS,SS>endif|fi		{ context = NIL;
				  return _ENDIF;
				}
<INITIAL,DS,SS>while		{ context = NUMBER;
				  return _WHILE;
				}
<INITIAL,DS,SS>while{W}*"("	{ context = NUMBER;
				  unput('[');
				  return _WHILE;
				}
<INITIAL,DS,SS>endwhile		{ context = NIL;
				  return _ENDWHILE;
				}
<INITIAL,DS,SS>repeat		{ context = NIL;
				  return _REPEAT;
				}
<INITIAL,DS,SS>until		{ context = NUMBER;
				  return _UNTIL;
				}
<INITIAL,DS,SS>until{W}*"("	{ context = NUMBER;
				  unput('[');
				  return _UNTIL;
				}
<INITIAL,DS,SS>for		{ context = NUMBER;
				  return _FOR;
				}
<INITIAL,DS,SS>for{W}*"("	{ context = NUMBER;
				  return _FORLP;
				}
<INITIAL,DS,SS>foreach		{ context = STRING;
				  return _FOREACH;
				}
<INITIAL,DS,SS>endfor|endforeach {
				  context = NIL;
				  return _ENDFOR;
				}
<INITIAL,DS,SS>break		{ context = NIL;
				  return _BREAK;
				}
<INITIAL,DS,SS>continue|cont	{ context = NIL;
				  return _CONTINUE;
				}
<INITIAL,DS,SS>switch		{ context = NIL;
				  return _SWITCH;
				}
<INITIAL,DS,SS>switch{W}*"("	{ context = NIL;
				  return _SWITCHLP;
				}
<INITIAL,DS,SS>case		{ context = NIL;
				  return _CASE;
				}
<INITIAL,DS,SS>default		{ context = NIL;
				  return _DEFAULT;
				}
<INITIAL,DS,SS>breakswitch|breaksw {
				  context = NIL;
				  return _BREAKSWITCH;
				}
<INITIAL,DS,SS>endswitch|endsw	{ context = NIL;
				  return _ENDSWITCH;
				}
<INITIAL,DS,SS>again		{ context = NIL;
				  return _AGAIN;
				}
<INITIAL,DS,SS>leave		{ context = NIL;
				  return _LEAVE;
				}
<INITIAL,DS,SS>redo		{ context = NIL;
				  return _REDO;
				}
<INITIAL,DS,SS>next		{ context = NIL;
				  return _NEXT;
				}
<INITIAL,DS,SS>skip		{ context = NUMBER;
				  return _SKIP;
				}
<INITIAL,DS,SS>last		{ context = NIL;
				  return _LAST;
				}
<INITIAL,DS,SS>pause		{ context = NUMBER;
				  return _PAUSE;
				}
<INITIAL,DS,SS>quit		{ context = NIL;
				  return _QUIT;
				}
<INITIAL,DS,SS>die		{ context = STRING;
				  return _DIE;
				}

<INITIAL,DS,SS>output		{ context = STRING;
				  return _OUTPUT;
				}
<INITIAL,DS,SS>mail		{ return _MAIL; }
<INITIAL,DS,SS>print		{ return _PRINT; }
<INITIAL,DS,SS>log		{ return _LOG; }
<INITIAL,DS,SS>piktlog		{ return _PIKTLOG; }
<INITIAL,DS,SS>syslog		{ return _SYSLOG; }

<INITIAL,DS,SS>set/[ \t\n]+(#|$){L}{DLU}*[ \t\n]+"=" {
				  isfakearray = TRUE;
				  context = NIL;
				  BEGIN(SS);
				  return _SET;
				}
<INITIAL,DS,SS>set		{ isfakearray = FALSE;
				  context = NIL;
				  BEGIN(SS);
				  return _SET;
				}
<INITIAL,DS,SS>unset|delete	{ context = NIL;
				  BEGIN(SS);
				  return _UNSET;
				}

<INITIAL,DS,SS>exec		{ context = STRING;
				  BEGIN(INITIAL);
				  return _EXEC;
				}
<INITIAL>wait			{ return _WAIT; }

<INITIAL,DS,SS>for{W}+${L}{DLU}*{W}+in{W}+#(members|keys|skeys|sortkeys|rkeys|reversekeys){W}*"(" {
				  /* for $k in #(members|...)( -->
				     foreach #(members|...)($k,
				   */
				  /* first, capture #(members|...) in s2 */
				  s2 = strchr(yytext, '#');
				  kk = 1;
				  while (isalpha((int)*(s2+kk))) {
					kk++;
				  }
				  *(s2+kk) = '\0';
				  /* next, capture $foo in s1 */
				  s1 = strchr(yytext, '$');
				  jj = 0;
				  while (! isspace((int)*(s1+jj))) {
					jj++;
				  }
				  *(s1+jj) = '\0';
				  /* write the replacement foreach string */
				  /* 11, because strlen("foreach ") plus
				     '(' plus ',' plus null char
				   */
				  yyt = (char *) mymalloc((unsigned)(11+jj+kk), _altlex_);
				  sprintf(yyt, "foreach %s(%s,", s2, s1);
				  /* push it back to the lexer input stream */
				  ii = 10+jj+kk;
				  while (ii) {
					unput(*(yyt+ii-1));
					ii--;
				  }
				}

<CS,DS>\"[^\n]+\"/[\n\t ]*	{ yytext[yyleng-1] = '\0';
				  BEGIN(INITIAL);
				  altlval.sval = mystrdup(yytext+1, _altlex_);
				  return _STR;
				}

<DS>"["+{W}*\"[^\"\n]+\"/{W}*"]" {
				  ii = 1;
				  while ((*(yytext+ii) == '\[') || isspace((int)*(yytext+ii)))
					ii++;
				  d = decompose(yytext+ii, outstr);
				  for (ii=d; ii>=0; ii--)
					unput(*(outstr+ii));
				  return '[';
				}
<DS>({D}+|$|$-{D}+)/[^\]0-9A-Za-z_] {
				  /* <DS>({D}{1,5}|$|$-{D}{1,5})/[^\]0-9A-Za-z_] */
				  BEGIN(INITIAL);
				  RETURN(_POSITION);
				}
<DS>"["({D}+|$|$-{D}+)"]" 	{ /* <DS>"["({D}{1,5}|$|$-{D}{1,5})"]" */
				  BEGIN(INITIAL);
				  RETURN(_POSITION);
				}
<DS>{D}+","({D}+|$)		{ BEGIN(INITIAL);
				  RETURN(_RANGE);
				}
<DS>"["{D}+","({D}+|$)"]"	{ BEGIN(INITIAL);
				  RETURN(_RANGE);
				}

<DS>(#|$){L}{DLU}*"[" 		{ nam = (char *) mymalloc((yyleng-1+3) * sizeof(char), _altlex_);
				  sprintf(nam, " %s ", yytext+1);
				  if (strstr(" inlin inline inputline rdlin rdline readline ", nam)) {
					alterror("syntax error");
				  }
				  free(nam);
				  yytext[yyleng-1] = '\0';
				  altlval.sval = mystrdup(yytext, _altlex_);
				  unput('[');
				  unput('[');
				  unput('[');
				  /*
				  context = (*altlval.sval == '#' ? NUMBER : STRING);
				   */
				  return (*altlval.sval == '#' ? _VARNUM : _VARSTR);
				}

<INITIAL,SS>"="			{ context = NIL;
				  BEGIN(INITIAL);
				  return yytext[0];
				}

<INITIAL,SS>(#|$){L}{DLU}*({W}*|"["[^\n]+"]"{W}*)("+="|"-="|"*="|"/="|"%="|"\*\*="|"<<="|">>="|"&="|"|="|"^="|".=") {
				  if (! strcmp(".=",yytext))
					context = STRING;
				  else
					context = NUMBER;
				  yyt = mystrdup(yytext, _altlex_);
				  yyl = yyleng;
				  ii = yyl - 2;
				  while (ii >= 0)
					unput(*(yyt+ii--));
				  unput(' ');
				  unput('=');
				  unput(' ');
				  cc = *(yyt+yyl-3);
				  ii = yyl - ((isspace((int)cc) || isalnum((int)cc)) ? 3 : 4);
				  while (ii >= 0)
					unput(*(yyt+ii--));
				  free(yyt);
				}

<INITIAL,DS,SS,FS>\'[^\'\n]*\'	{ context = STRING;
				  yytext[yyleng-1] = '\0';
				  altlval.sval = mystrdup(yytext+1, _altlex_);
				  return _STR;
				}

<FS>[A-Z][0-9A-Z]*		{ nam = (char *) mymalloc((yyleng+2) * sizeof(char), _altlex_);
				  sprintf(nam, "?%s", yytext);
				  altlval.sval = mystrdup(nam, _altlex_);
				  free(nam);
				  return _FLHNDL;
				}

<INITIAL,SS,FS>${D}+		{ /* <INITIAL,SS,FS>${D}{1,5} */
				  altlval.sval = mystrdup("$", _altlex_);
				  yyt = mystrdup(yytext, _altlex_);
				  ii = yyleng-1;
				  unput(']');
				  while (ii > 0)
					unput(*(yyt+ii--));
				  free(yyt);
				  unput('[');
				  unput('[');
				  unput('[');
				  context = STRING;
				  return _VALSTR;
				}

<INITIAL,DS,SS,FS>(#|$){L}{DLU}* {
				  altlval.sval = mystrdup(yytext, _altlex_);
				  FAKEARRAY;
				  context = (*altlval.sval == '#' ? NUMBER : STRING);
				  return (*altlval.sval == '#' ? _VARNUM : _VARSTR);
				}

<INITIAL,SS,FS>$"["		{ altlval.sval = mystrdup("$", _altlex_);
				  unput('[');
				  unput('[');
				  unput('[');
				  context = STRING;
				  return _VALSTR;
				}

<INITIAL,SS,FS>(#|$){L}{DLU}*{W}*"[" {
				  nam = (char *) mymalloc((yyleng-1+3) * sizeof(char), _altlex_);
				  sprintf(nam, " %s ", yytext+1);
				  if (strstr(" inlin inline inputline rdlin rdline readline ", nam)) {
					alterror("syntax error");
				  }
				  free(nam);
				  ii = yyleng-2;
				  while (isspace((int)*(yytext+ii)))
					ii--;
				  *(yytext+ii+1) = '\0';
				  altlval.sval = mystrdup(yytext, _altlex_);
				  unput('[');
				  unput('[');
				  unput('[');
				  /*
				  context = (*altlval.sval == '#' ? NUMBER : STRING);
				   */
				  return (*altlval.sval == '#' ? _VARNUM : _VARSTR);
				}

<INITIAL,FS>(%|@){L}{DLU}*/{W}*("+"|"-"|"*"|"/"|"%"|"**"|"& "|"| "|"^"|"<<"|">>"|"=="|">"|"<"|">="|"<="|"!>"|"!<"|"!="|"<>"|"]"){W}* {
				  altlval.sval = mystrdup(yytext, _altlex_);
				  FAKEARRAY;
				  context = NUMBER;
				  return (*altlval.sval == '%' ? _VARHSTNUM : _VARPRCNUM);
				}

<INITIAL,FS>(%|@){L}{DLU}*/{W}*"["[^\[\]]+"]"{W}*("+"|"-"|"*"|"/"|"%"|"**"|"& "|"| "|"^"|"<<"|">>"|"=="|">"|"<"|">="|"<="|"!>"|"!<"|"!="|"<>"|"]"){W}* {
				  context = NUMBER;
				  RETURN(yytext[0] == '%' ? _VARHSTNUM : _VARPRCNUM);
				}

<INITIAL,FS>(%|@){L}{DLU}*/{W}*("."|"=~"|"=~~"|"!~"|"!~~"){W}* {
				  altlval.sval = mystrdup(yytext, _altlex_);
				  FAKEARRAY;
				  context = STRING;
				  return (*altlval.sval == '%' ? _VARHSTSTR : _VARPRCSTR);
				}

<INITIAL,FS>(%|@){L}{DLU}*/{W}*"["[^\[\]]+"]"{W}*("."|"=~"|"=~~"|"!~"|"!~~"){W}* {
				  context = STRING;
				  RETURN(yytext[0] == '%' ? _VARHSTSTR : _VARPRCSTR);
				}

<INITIAL,FS>(%|@){L}{DLU}*/{W}+(eq|ne|lt|gt|le|ge){W}* {
				  altlval.sval = mystrdup(yytext, _altlex_);
				  FAKEARRAY;
				  context = STRING;
				  return (*altlval.sval == '%' ? _VARHSTSTR : _VARPRCSTR);
				}

<INITIAL,FS>(%|@){L}{DLU}*/{W}*"["[^\[\]]+"]"{W}+(eq|ne|lt|gt|le|ge){W}* {
				  context = STRING;
				  RETURN(yytext[0] == '%' ? _VARHSTSTR : _VARPRCSTR);
				}

<INITIAL,FS>(%|@){L}{DLU}*{W}*.	{ if (yytext[yyleng-1] == '\[')
					isarray = TRUE;
				  else
				  	isarray = FALSE;
				  altlval.sval = mystrdup(yytext, _altlex_);
				  ii = yyleng-1;
				  unput(*(altlval.sval+ii--));
				  while (isspace((int)*(altlval.sval+ii)))
					ii--;
				  *(altlval.sval+ii+1) = '\0';
				  if (! isarray) {
					FAKEARRAY;
				  }
				  if (! prescan) {
					switch (vardattype(altlval.sval)) {
						case STRING:
							return (*altlval.sval == '%' ? _VARHSTSTR : _VARPRCSTR);
							break;
						case NUMBER:
							return (*altlval.sval == '%' ? _VARHSTNUM : _VARPRCNUM);
							break;
						case ERROR:
						default:
							free(altlval.sval);
							altlval.sval = mystrdup(err, _altlex_);
							return _STR;
							break;
					}
				  } else {
					switch (context) {
						case NUMBER:
							return (*altlval.sval == '%' ? _VARHSTNUM : _VARPRCNUM);
							break;
						case STRING:
						case NIL:
						default:
							return (*altlval.sval == '%' ? _VARHSTSTR : _VARPRCSTR);
							break;
					}
				  }
				}

<INITIAL,DS,SS,FS>"["/{W}*($|\') {
				  context = STRING;
				  return yytext[0];
				}
<INITIAL,DS,SS,FS>"["		{ context = NUMBER;
				  return yytext[0];
				}
<INITIAL,DS,SS,FS>"[[["		{ return '['; }
<INITIAL,DS,SS,FS>{D}+		{ /* <INITIAL,DS,SS,FS>{D}{1,5}	*/
				  context = NUMBER;
				  RETURN(_INT);
				}
<INITIAL,DS,SS,FS>"]"		{ return yytext[0]; }

<INITIAL,SS,FS>("+"|"-"|"~")	{ context = NUMBER;
				  return yytext[0];
				}

<INITIAL,SS,FS>("*"|"/"|"%"|"**"|"&"|"|"|"^"|"<<"|">>") {
				  context = NUMBER;
				  RETURN(_NOP);
				}

<INITIAL,SS,FS>"."		{ context = STRING;
				  RETURN(_SOP);
				}

<INITIAL,SS,FS>(-e|-z|-f|-d|-c|-b|-p|-l|-S) {
				  context = STRING;
				  RETURN(_FOP);
				}

<INITIAL,SS,FS>("=="|">"|"<"|">="|"<="|"!>"|"!<"|"!="|"<>") {
				  context = NUMBER;
				  RETURN(_NREL);
				}
<INITIAL,SS,FS>("=~"|"=~~"|"!~"|"!~~"|eq|ne|lt|gt|le|ge) {
				  context = STRING;
				  RETURN(_SREL);
				}

<INITIAL,SS,FS>(({D}+)|({D}*\.{D}+)|((({D}+)|({D}*\.{D}+))\/(({D}+)|({D}*\.{D}+))))%* {
				  context = NUMBER;
				  RETURN(_NUM);
				}

<INITIAL,SS,FS>(#|$){L}{DLU}*/{W}*"(" {
				  context = (yytext[0] == '#' ? NUMBER : STRING);
				  BEGIN(FS);
				  RETURN(yytext[0] == '#' ? _FNNUM : _FNSTR);
				}
<SS,FS>"("			{ context = NIL;
				  fn++;
				  return '(';
				}
<SS,FS>")"			{ context = NIL;
				  fn--;
				  if (! fn)
					BEGIN(INITIAL);
				  return ')';
				}

<INITIAL,SS,FS>\"[^\n]*\"	{ if ((d = decompose(yytext, outstr)) > 0) {
					for (ii=d; ii>=0; ii--)
						unput(*(outstr+ii));
				  } else {
					*(outstr+strlen(outstr)-1) = '\0';
					altlval.sval = mystrdup(outstr+1, _altlex_);
					return _STRVAR;
				  }
				}

<INITIAL>"("[\n\t ]*[^\n\t ]+[\n\t ]*(=~|=~~|!~|!~~|eq|ne|lt|gt|le|ge)[\n\t ]*[^\n\t ]+[\n\t ]*")"[\n\t ]* { 
				  context = STRING;
				  yyless(1);
				  return '[';
				}

<INITIAL>"("([-+*/%$#@~^><)(,.0-9A-Za-z_\n\t\[\] ]|"&"[^&]|"|"[^|])+ {
				  if (condparen(yytext, context)) {
					context = NUMBER;
					yyless(1);
					return '[';
				  } else {
					context = NUMBER;
					yyless(1);
					return yytext[0];
				  }
				}

<INITIAL>")"			{ return yytext[0]; }

<INITIAL,SS,FS>"!"		{ context = NUMBER;
				  return _NOT;
				}
<INITIAL,SS,FS>"&&"		{ context = NUMBER;
				  return _AND;
				}
<INITIAL,SS,FS>"||"		{ context = NUMBER;
				  return _OR;
				}

<*>","				{ return yytext[0]; }

<*>{W}				{ ; }
<*>\n				{ lineno++; }

<*>.				{ return yytext[0]; }

%%

/* alt error handler
 */
void
alterror (const char *message)
{
	static int l = 0;

	if (l != lineno)
		l = lineno;
	logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in .alt file, at or near line %d, %s (in altparse()): @%s@", lineno, message, yytext ? yytext : "");
	BEGIN(INITIAL);
	return;
}
