/*
   scan.c -- PIKT input scanning routines

   $Header: /home/cvs/ucgsb/pikt/src/piktd/scan.c,v 1.36 2000/06/04 00:58:45 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2000 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
*/

#include "../lib/common.h"

#include "piktd.h"

int inlen;

signed char *valscn[MAXVALSCN+1];

int issep(int, char *);

/* dequote() the current input line; depending on whether looking for fields,
   columns, or regular expressions, scan the current input line, assigning
   values to dat variables
 */
int
scaninput (int postype)
{
	static char _scaninput_[] = "scaninput";

	if ((inlen = dequote(inlinnow)-1) <= 0)  /* subtract one for eol */
		return ERROR;
	switch (postype) {
		case FIELDS:
			freevalscn();
			if ((matches = scanfields(0, inlinnow, SEPARATORS, FALSE)) != DATCNT) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scaninput(), %s, data mismatch in line: %s", ALMNAM, inlinnow);
				return ERROR;
			}
			break;
		case COLUMNS:
			freevalscn();
			if ((matches = scancolumns(inlinnow)) != DATCNT) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scaninput(), %s, data mismatch in line: %s", ALMNAM, inlinnow);
				return ERROR;
			}
			break;
		case REGEXPR:
			/* done in regexp()
			freevalscn();
			 */
			regexp(0, inlinnow, REGEXP, REG_EXTENDED, FALSE);
			break;
		default:	/* no dat statements,
				   postype defaults to 0
				 */
			freevalscn();
			matches = scanfields(0, inlinnow, SEPARATORS, FALSE);
			break;
	}
	/* chop off linefeed from end of input line */
	if (*(inlinnow+inlen) == '\n')
		*(inlinnow+inlen) = '\0';
	if (postype != REGEXPR)
		valscn[0] = mystrdup(inlinnow, _scaninput_);
	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "input line is: @%s@", inlinnow);
		dumpvals(valnow);
	}
	return 0;
}

/* scan the current input line, or the string passed in the #split()
   function; break down the line into component parts, with white space
   as the default separator, or using the specified separator char(s);
   assign scanned vals to the appropriate var index;
   scanfields() is for dat statements of the form "dat #foo 1",
   or #split() invocations of the form "#split($rdlin)"
 */
int
scanfields (int vrndx, char *line, char *separators, int isparsefn)
{
	static char _scanfields_[] = "scanfields";
  	static char *buf;
	unsigned char s;
	register int i, j, k;
	int col;
	int f, fn;
	register char c;
	int fields;
	int done;
	int nspsep;
	spc *sp, *initsp;
	int direction;
	vl *vlp;

	nspsep = FALSE;
	fields = 0;
	s = OUTDAT;
	i = 0;
	col = 0;
	f = 1;
	buf = mystrdup(line, _scanfields_);
	while ((c = *(line+col))) {
		if ((isspace((int)c) && issep(c, separators)) || (c == '\n')) {
			switch (s) {
				case INDAT:
				case BEGINDAT:
					s = ENDDAT;
					break;
				default:
					s = OUTDAT;
					break;
			}
		} else if (issep(c, separators)) {
			nspsep = TRUE;
			/* in all cases, we set s to ENDDAT! */
			switch (s) {
				/*
				case INDAT:
					s = ENDDAT;
					break;
				case BEGINDAT:
					s = ENDDAT;
					break;
				case ENDDAT:
					s = ENDDAT;
					break;
				case OUTDAT:
					s = ENDDAT;
					break;
				*/
				default:
					s = ENDDAT;
					break;
			}
		} else {	/* is not eol, space, tab, or sep */
			switch (s) {
				case OUTDAT:
				case ENDDAT:
					s = BEGINDAT;
					break;
				default:
					s = INDAT;
					break;
			}
		}
		switch (s) {	
			case BEGINDAT:
			case INDAT:
				*(buf+i++) = c;
				break;
			case ENDDAT:
				*(buf+i) = '\0';
				FREE(valscn[f]);
				valscn[f] = mystrdup(buf, _scanfields_);
				for (k=0; k<=MAXARYDIM; k++)
					arndx[k] = 0L;
				arndx[1] = (signed long) f;
				if (vrndx) {
					if ((vlp = lookupvl(valnow, vrndx, arndx, (char *)NULL, ARYTYP(vrndx)))) {
						FREE(vlp->str);
					} else {
						vlp = regvl(valnow, vrndx, arndx, (char *)NULL, ARYTYP(vrndx));
					}
					vlp->str = mystrdup(buf, _scanfields_);
					errno = 0;
					vlp->num = (double) atof(vlp->str);
					if (errno)
						vlp->num = ERROR;
				}
				i = 0;
				f++;
				break;
			case OUTDAT:
				if (c == '\n') {
					*(buf+i) = '\0';
				}
				break;
		}
		col++;
	}
	if ((s == ENDDAT) || ! strcmp(line, "\n"))
		f--;
	if (! valscn[f]) {
		/* if we are parsing $command() output, which doesn't end in a
		   linefeed, we have to terminate the last buf string here
		 */
		if (isparsefn)
			*(buf+i) = '\0';
		if (isparsefn || nspsep) {
			valscn[f] = mystrdup(buf, _scanfields_);
			for (k=0; k<=MAXARYDIM; k++)
				arndx[k] = 0L;
			arndx[1] = (signed long) f;
			if (vrndx) {
				if ((vlp = lookupvl(valnow, vrndx, arndx, (char *)NULL, ARYTYP(vrndx)))) {
					FREE(vlp->str);
				} else {
					vlp = regvl(valnow, vrndx, arndx, (char *)NULL, ARYTYP(vrndx));
				}
				vlp->str = mystrdup(buf, _scanfields_);
				errno = 0;
				vlp->num = (double) atof(vlp->str);
				if (errno)
					vlp->num = ERROR;
			}
		} else
			f--;
	}
	if ((DATCNT == 0) || isparsefn)	{ /* no dat statements */
		free(buf);
		return f;
	}
	if (f < MAXDATNDX) {	/* f is # of dat valscn parsed */
		free(buf);
		return ERROR;
	}
	done = FALSE;
	for (i=1; i<=VRNDXHI; i++) {
		if (done) break;
		if (VALTYP(i) & FIXEDVAL) {
			direction = DIRECTION(i);
			for (j=0; j<=9; j++) {
				if (done) break;
				sp = initsp = (alarmcurr->vars[i])->datspc[j];
				while (sp) {
					if (done) break;
					fn = sp->fldnum;
					if (fn == 0)
						continue;
					else if (fn < 0)
						fn = f+fn+1;
					vlp = regvl(valnow, i, sp->ndx, sp->key, ARYTYP(i));
					if (sp->trim == BOTH)
						vlp->str = trim(valscn[fn],BOTH);
					else
						vlp->str = mystrdup(valscn[fn], _scanfields_);
					errno = 0;
					vlp->num = (double) atof(vlp->str);
					if (errno)
						vlp->num = ERROR;
					if (++fields >= DATCNT)
						done = TRUE;
					if (direction == FORWARD)
						sp = sp->next;
					else
						sp = sp->prev;
					if (sp == initsp)
						break;
				}
			}
		}
	}
	free(buf);
	return fields;
}

/* scan the current input line; break down the line into component parts,
   according to their column position; assign scanned vals to the appropriate
   var index;
   scancolumns() is for dat statements of the form "dat #foo 10,15"
 */
int
scancolumns (char *line)
{
	static char _scancolumns_[]= "scancolumns";
	static char *buf;
	unsigned char s;
	register int i;
	int col;
	int f;
	register char c;
	ky *scvar;
	int begincol, endcol;
	spc *sp;
	vl *vlp;

	i = 0;
	col = 0;
	f = 1;
	buf = mystrdup(line, _scancolumns_);
	scvar = SCVARS;
	sp = lookupspc(scvar->vrndx, scvar->arndx, scvar->arkey);
	begincol = sp->begincol;
	endcol = sp->endcol;
	if (! VARNAM(f)) {
		free(buf);
		return 0;
	}
	while ((c = *(line+col))) {
		col++;
		if (c == '\n')
			s = ENDDAT;
		else if (col < begincol)
			s = OUTDAT;
		else if (col == begincol)
			s = BEGINDAT;
		else if (col <= endcol)
			s = INDAT;
		else if (col == endcol+1)
			s = ENDDAT;
		else
			s = OUTDAT;
		switch (s) {
			case BEGINDAT:
			case INDAT:
				*(buf+i++) = c;
				break;
			case ENDDAT:
				*(buf+i) = '\0';
				valscn[f] = mystrdup(buf, _scancolumns_);
				vlp = regvl(valnow, scvar->vrndx, scvar->arndx, scvar->arkey, ARYTYP(scvar->vrndx));
				if (sp->trim == BOTH)
					vlp->str = trim(valscn[f],BOTH);
				else
					vlp->str = mystrdup(valscn[f], _scancolumns_);
				errno = 0;
				vlp->num = (double) atof(vlp->str);
				if (errno)
					vlp->num = ERROR;
				i = 0;
				if (++f > DATCNT) {
					free(buf);
					return --f;
				}
				scvar = scvar->next;
				if (! scvar) {
					free(buf);
					return --f;
				}
				sp = lookupspc(scvar->vrndx, scvar->arndx, scvar->arkey);
				begincol = sp->begincol;
				endcol = sp->endcol;
				if (c != '\n')
					col--;
				break;
			case OUTDAT:
				break;
		}
	}
	free(buf);
	return f;
}

/* is the given char in the separators list (defaults to space (" "))?
 */
int
issep (int c, char *seps)
{
	register int i;
	register char sc;

	if (! seps)	/* should never be true */
		return FALSE;
	else {
		i = 0;
		while ((sc = *(seps+i))) {
			if (sc == c)
				return sc;
			i++;
		}
	}
	return FALSE;
}

/* strip backslash escape chars from the given scan command; this is to
   ensure consistency of treatment with escape processing done in begin,
   end, and rules sections, where backslash escapes are dealt with by
   the decompose() function
 */
char *
stripesc (char *input)
{
	static char output[MAXLINE+1];
	char c;
	register int i, j;
	register int escape;

	escape = FALSE;
	i = j = 0;
	while ((c = *(input+i))) {
		i++;
		if ((c == '\\') && ! escape) {
			escape = TRUE;
			continue;
		}
		output[j++] = c;
		escape = FALSE;
	}
	output[j] = '\0';
	return output;
}

/* for input str, replace all double quotes (") with DQTPROXY,
   and all single quotes (') with SQTPROXY, returning strlen
 */
int
dequote (signed char *str)
{
	register int i;
	register signed char c;

	i = 0;
	while ((c = *(str+i))) {
		if (c == '\"')
			*(str+i) = DQTPROXY;
		else if (c == '\'')
			*(str+i) = SQTPROXY;
		i++;
	}
	return i;
}

/* for input str, replace all DQTPROXY with double quotes ("),
   and all SQTPROXY with single quotes ('), returning strlen
 */
int
requote (signed char *str)
{
	register int i;
	register signed char c;

	if (! str)
		return ERROR;
	i = 0;
	while ((c = *(str+i))) {
		if (c == DQTPROXY)
			*(str+i) = '\"';
		else if (c == SQTPROXY)
			*(str+i) = '\'';
		i++;
	}
	return i;
}

/*  garbage collect valscn array
 */
void
freevalscn (void)
{
	register int i;

	for (i=0; i<=matches; i++) {
		if (valscn[i]) {
			free(valscn[i]);
			valscn[i] = (signed char *) NULL;
		}
	}
}
